# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

from power_grid_model import CalculationType
from power_grid_model.utils import import_input_data
from power_grid_model.validation import errors_to_string, validate_input_data

input_file = Path("../tests/data/state_estimation/dummy-test-sym/input.json")

input_data = import_input_data(json_file=input_file)

input_errors = validate_input_data(
    input_data=input_data, calculation_type=CalculationType.state_estimation, symmetric=True
)

print(errors_to_string(input_errors, name=str(input_file), details=True))



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Error classes
"""

import re
from abc import ABC
from collections.abc import Iterable
from enum import Enum
from typing import Any

from power_grid_model import ComponentType
from power_grid_model._core.dataset_definitions import DatasetType

_MIN_FIELDS = 2
_MIN_COMPONENTS = 2


class ValidationError(ABC):
    """
    The Validation Error is an abstract base class which should be extended by all validation errors. It supplies
    three public member variables: component, field and ids; storing information about the origin of the validation
    error. Error classes can extend the public members. For example:

        NotBetweenError(ValidationError):
            component = 'vehicle'
            field = 'direction'
            id = [3, 14, 15, 92, 65, 35]
            ref_value = (-3.1416, 3.1416)

    For convenience, a human readable representation of the error is supplied using the str() function.
    I.e. print(str(error)) will print a human readable error message like:

        Field `direction` is not between -3.1416 and 3.1416 for 6 vehicles

    """

    component: ComponentType | list[ComponentType] | None = None
    """
    The component, or components, to which the error applies.
    """

    field: str | list[str] | list[tuple[ComponentType, str]] | None = None
    """
    The field, or fields, to which the error applies. A field can also be a tuple (component, field) when multiple
    components are being addressed.
    """

    ids: list[int] | list[tuple[ComponentType, int]] | None = None
    """
    The object identifiers to which the error applies. A field object identifier can also be a tuple (component, id)
    when multiple components are being addressed.
    """

    _message: str = "An unknown validation error occurred."

    _delimiter: str = " and "

    __hash__ = None  # type: ignore[assignment]

    @property
    def component_str(self) -> str:
        """
        A string representation of the component to which this error applies
        """
        if self.component is None:
            return str(None)
        if isinstance(self.component, list):
            return "/".join(component.value for component in self.component)
        return self.component.value

    @property
    def field_str(self) -> str:
        """
        A string representation of the field to which this error applies
        """

        def _unpack(field: str | tuple[ComponentType, str]) -> str:
            if isinstance(field, str):
                return f"'{field}'"
            return ".".join(field)

        if isinstance(self.field, list):
            return self._delimiter.join(_unpack(field) for field in self.field)
        return _unpack(self.field) if self.field else str(self.field)

    def get_context(self, id_lookup: list[str] | dict[int, str] | None = None) -> dict[str, Any]:
        """
        Returns a dictionary that supplies (human readable) information about this error. Each member variable is
        included in the dictionary. If a function {field_name}_str() exists, the value is overwritten by that function.

        Args:
            id_lookup: A list or dict (int->str) containing textual object ids
        """
        context = self.__dict__.copy()
        if id_lookup:
            if isinstance(id_lookup, list):
                id_lookup = dict(enumerate(id_lookup))
            context["ids"] = (
                {i: id_lookup.get(i[1] if isinstance(i, tuple) else i) for i in self.ids} if self.ids else set()
            )
        for key in context:
            if hasattr(self, key + "_str"):
                context[key] = str(getattr(self, key + "_str"))
        return context

    def __str__(self) -> str:
        n_objects = len(self.ids) if self.ids else 0
        context = self.get_context()
        context["n"] = n_objects
        context["objects"] = context.get("component", "object")
        if n_objects != 1:
            context["objects"] = re.sub(r"([a-z_]+)", r"\1s", context["objects"])
        return self._message.format(**context).strip()

    def __repr__(self) -> str:
        context = " ".join(f"{key}={value}" for key, value in self.get_context().items())
        return f"<{type(self).__name__}: {context}>"

    def __eq__(self, other):
        return (
            type(self) is type(other)
            and self.component == other.component
            and self.field == other.field
            and self.ids == other.ids
        )


class SingleFieldValidationError(ValidationError):
    """
    Base class for an error that applies to a single field in a single component
    """

    _message = "Field {field} is not valid for {n} {objects}."
    component: ComponentType
    field: str
    ids: list[int] | None

    def __init__(self, component: ComponentType, field: str, ids: Iterable[int] | None):
        """
        Args:
            component: Component name
            field: Field name
            ids: List of component IDs (not row indices)
        """
        self.component = component
        self.field = field
        self.ids = sorted(ids) if ids is not None else None


class MultiFieldValidationError(ValidationError):
    """
    Base class for an error that applies to multiple fields in a single component
    """

    _message = "Combination of fields {field} is not valid for {n} {objects}."
    component: ComponentType
    field: list[str]
    ids: list[int]

    def __init__(self, component: ComponentType, fields: list[str], ids: list[int]):
        """
        Args:
            component: Component name
            fields: List of field names
            ids: List of component IDs (not row indices)
        """
        self.component = component
        self.field = sorted(fields)
        self.ids = sorted(ids)

        if len(self.field) < _MIN_FIELDS:
            raise ValueError(f"{type(self).__name__} expects at least 2 fields: {self.field}")


class MultiComponentValidationError(ValidationError):
    """
    Base class for an error that applies to multiple components, and, subsequently, multiple fields.
    Even if both fields have the same name, they are considered to be different fields and notated as such.
    E.g. the two fields `id` fields of the `node` and `line` component: [('node', 'id'), ('line', 'id')].
    """

    component: list[ComponentType]
    field: list[tuple[ComponentType, str]]
    ids: list[tuple[ComponentType, int]]
    _message = "Fields {field} are not valid for {n} {objects}."

    def __init__(self, fields: list[tuple[ComponentType, str]], ids: list[tuple[ComponentType, int]]):
        """
        Args:
            fields: List of field names, formatted as tuples (component, field)
            ids: List of component IDs (not row indices), formatted as tuples (component, id)
        """
        self.component = sorted(set(component for component, _ in fields), key=str)
        self.field = sorted(fields)
        self.ids = sorted(ids)

        if len(self.field) < _MIN_FIELDS:
            raise ValueError(f"{type(self).__name__} expects at least {_MIN_FIELDS} fields: {self.field}")
        if len(self.component) < _MIN_COMPONENTS:
            raise ValueError(f"{type(self).__name__} expects at least {_MIN_COMPONENTS} components: {self.component}")


class NotIdenticalError(SingleFieldValidationError):
    """
    The value is not unique within a single column in a dataset
    E.g. When two nodes share the same id.
    """

    _message = "Field {field} is not unique for {n} {objects}: {num_unique} different values."
    values: list[Any]
    unique: set[Any]
    num_unique: int

    def __init__(self, component: ComponentType, field: str, ids: Iterable[int], values: list[Any]):
        super().__init__(component, field, ids)
        self.values = values
        self.unique = set(self.values)
        self.num_unique = len(self.unique)


class NotUniqueError(SingleFieldValidationError):
    """
    The value is not unique within a single column in a dataset
    E.g. When two nodes share the same id.
    """

    _message = "Field {field} is not unique for {n} {objects}."


class MultiComponentNotUniqueError(MultiComponentValidationError):
    """
    The value is not unique between multiple columns in multiple components
    E.g. When a node and a line share the same id.
    """

    _message = "Fields {field} are not unique for {n} {objects}."


class InvalidValueError(SingleFieldValidationError):
    """
    The value is not a valid value in the supplied list of supported values.
    E.g. an enum value that is not supported for a specific feature.
    """

    _message = "Field {field} contains invalid values for {n} {objects}."
    values: list
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, component: ComponentType, field: str, ids: list[int], values: list):
        super().__init__(component, field, ids)
        self.values = values

    @property
    def values_str(self) -> str:
        """
        A string representation of the field to which this error applies.
        """
        return ",".join(v.name if isinstance(v, Enum) else v for v in self.values)

    def __eq__(self, other):
        return super().__eq__(other) and self.values == other.values


class InvalidEnumValueError(SingleFieldValidationError):
    """
    The value is not a valid value in the supplied enumeration type.
    E.g. a sym_load has a non existing LoadGenType.
    """

    _message = "Field {field} contains invalid {enum} values for {n} {objects}."
    enum: type[Enum] | list[type[Enum]]
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, component: ComponentType, field: str, ids: list[int], enum: type[Enum] | list[type[Enum]]):
        super().__init__(component, field, ids)
        self.enum = enum

    @property
    def enum_str(self) -> str:
        """
        A string representation of the field to which this error applies.
        """
        if isinstance(self.enum, list):
            return ",".join(e.__name__ for e in self.enum)

        return self.enum.__name__

    def __eq__(self, other):
        return super().__eq__(other) and self.enum == other.enum


class SameValueError(MultiFieldValidationError):
    """
    The value of two fields is equal.
    E.g. A line has the same from_node as to_node.
    """

    _message = "Same value for {field} for {n} {objects}."


class NotBooleanError(SingleFieldValidationError):
    """
    Invalid boolean value. Boolean fields don't really exist in our data structure, they are 1-byte signed integers and
    should contain either a 0 (=False) or a 1 (=True).
    """

    _message = "Field {field} is not a boolean (0 or 1) for {n} {objects}."


class MissingValueError(SingleFieldValidationError):
    """
    A required value was missing, i.e. NaN.
    """

    _message = "Field {field} is missing for {n} {objects}."


class IdNotInDatasetError(SingleFieldValidationError):
    """
    An object identifier does not exist in the original data.
    E.g. An update data set contains a record for a line that doesn't exist in the input data set.
    """

    _message = "ID does not exist in {ref_dataset} for {n} {objects}."
    ref_dataset: DatasetType
    __hash__ = None  # type: ignore[assignment]

    def __init__(self, component: ComponentType, ids: list[int], ref_dataset: DatasetType):
        super().__init__(component=component, field="id", ids=ids)
        self.ref_dataset = ref_dataset

    def __eq__(self, other):
        return super().__eq__(other) and self.ref_dataset == other.ref_dataset


class InvalidIdError(SingleFieldValidationError):
    """
    An object identifier does not refer to the right type of object.
    E.g. An line's from_node refers to a source.

    Filters can have been applied to check a subset of the records. E.g. This error may apply only to power_sensors
    that are said to be connected to a source (filter: measured_terminal_type=source). This useful to spot validation
    mistakes, due to ambiguity.

    E.g. when a sym_power_sensor is connected to a load, but measured_terminal_type is accidentally set to 'source',
    the error is:

    "Field `measured_object` does not contain a valid source id for 1 sym_power_sensor. (measured_terminal_type=source)"

    """

    _message = "Field {field} does not contain a valid {ref_components} id for {n} {objects}. {filters}"
    ref_components: list[ComponentType]
    __hash__ = None  # type: ignore[assignment]

    def __init__(
        self,
        component: ComponentType,
        field: str,
        ids: list[int] | None = None,
        ref_components: ComponentType | list[ComponentType] | None = None,
        filters: dict[str, Any] | None = None,
    ):
        super().__init__(component=component, field=field, ids=ids)
        ref_components = ref_components if ref_components is not None else []
        self.ref_components = [ref_components] if isinstance(ref_components, (str, ComponentType)) else ref_components
        self.filters = filters if filters else None

    @property
    def ref_components_str(self):
        """
        A string representation of the components to which this error applies
        """
        return "/".join(self.ref_components)

    @property
    def filters_str(self):
        """
        A string representation of the filters that have been applied to the data to which this error refers
        """
        if not self.filters:
            return ""
        filters = ", ".join(f"{k}={v.name if isinstance(v, Enum) else v}" for k, v in self.filters.items())
        return f"({filters})"

    def __eq__(self, other):
        return super().__eq__(other) and self.ref_components == other.ref_components and self.filters == other.filters


class TwoValuesZeroError(MultiFieldValidationError):
    """
    A record has a 0.0 value in two fields at the same time.
    E.g. A line's `r1`, `x1` are both 0.
    """

    _message = "Fields {field} are both zero for {n} {objects}."


class ComparisonError(SingleFieldValidationError):
    """
    Base class for comparison errors.
    E.g. A transformer's `i0` is not greater or equal to it's `p0` divided by it's `sn`
    """

    _message = "Invalid {field}, compared to {ref_value} for {n} {objects}."

    RefType = int | float | str | tuple[int | float | str, ...]

    __hash__ = None  # type: ignore[assignment]

    def __init__(self, component: ComponentType, field: str, ids: list[int], ref_value: "ComparisonError.RefType"):
        super().__init__(component, field, ids)
        self.ref_value = ref_value

    @property
    def ref_value_str(self):
        """
        A string representation of the reference value. E.g. 'zero', 'one', 'field_a and field_b' or '123'.
        """
        if isinstance(self.ref_value, tuple):
            return self._delimiter.join(map(str, self.ref_value))
        if self.ref_value == 0:
            return "zero"
        if self.ref_value == 1:
            return "one"
        return str(self.ref_value)

    def __eq__(self, other):
        return super().__eq__(other) and self.ref_value == other.ref_value


class NotGreaterThanError(ComparisonError):
    """
    The value of a field is not greater than a reference value or expression.
    """

    _message = "Field {field} is not greater than {ref_value} for {n} {objects}."


class NotGreaterOrEqualError(ComparisonError):
    """
    The value of a field is not greater or equal to a reference value or expression.
    """

    _message = "Field {field} is not greater than (or equal to) {ref_value} for {n} {objects}."


class NotLessThanError(ComparisonError):
    """
    The value of a field is not less than a reference value or expression.
    """

    _message = "Field {field} is not smaller than {ref_value} for {n} {objects}."


class NotLessOrEqualError(ComparisonError):
    """
    The value of a field is not smaller or equal to a reference value or expression.
    """

    _message = "Field {field} is not smaller than (or equal to) {ref_value} for {n} {objects}."


class NotBetweenError(ComparisonError):
    """
    The value of a field is not between two a reference values or expressions (exclusive).
    """

    _message = "Field {field} is not between {ref_value} for {n} {objects}."


class NotBetweenOrAtError(ComparisonError):
    """
    The value of a field is not between two a reference values or expressions (inclusive).
    """

    _message = "Field {field} is not between (or at) {ref_value} for {n} {objects}."


class InfinityError(SingleFieldValidationError):
    """
    The value of a field is infinite.
    """

    _message = "Field {field} is infinite for {n} {objects}."


class TransformerClockError(MultiFieldValidationError):
    """
    Invalid clock number.
    """

    _message = (
        "Invalid clock number for {n} {objects}. "
        "If one side has wye winding and the other side has not, the clock number should be odd. "
        "If either both or none of the sides have wye winding, the clock number should be even."
    )


class FaultPhaseError(MultiFieldValidationError):
    """
    The fault phase does not match the fault type.
    """

    _message = "The fault phase is not applicable to the corresponding fault type for {n} {objects}."


class PQSigmaPairError(MultiFieldValidationError):
    """
    The combination of p_sigma and q_sigma is not valid. They should be both present or both absent.
    """

    _message = "The combination of p_sigma and q_sigma is not valid for {n} {objects}."


class InvalidAssociatedEnumValueError(MultiFieldValidationError):
    """
    The value is not a valid value in combination with the other specified attributes.
    E.g. When a transformer tap regulator has a branch3 control side but regulates a transformer.
    """

    _message = "The combination of fields {field} results in invalid {enum} values for {n} {objects}."
    enum: type[Enum] | list[type[Enum]]
    __hash__ = None  # type: ignore[assignment]

    def __init__(
        self,
        component: ComponentType,
        fields: list[str],
        ids: list[int],
        enum: type[Enum] | list[type[Enum]],
    ):
        """
        Args:
            component: Component name
            fields: List of field names
            ids: List of component IDs (not row indices)
            enum: The supported enum values
        """
        super().__init__(component, fields, ids)
        self.enum = enum

    @property
    def enum_str(self) -> str:
        """
        A string representation of the field to which this error applies.
        """
        if isinstance(self.enum, list):
            return ",".join(e.__name__ for e in self.enum)

        return self.enum.__name__

    def __eq__(self, other):
        return super().__eq__(other) and self.enum == other.enum


class UnsupportedMeasuredTerminalType(InvalidValueError):
    """
    The measured terminal type is not a supported value.

    Supported values are in the supplied list of values.
    """

    _message = "measured_terminal_type contains unsupported values for {n} {objects}."


class MixedCurrentAngleMeasurementTypeError(MultiFieldValidationError):
    """
    Mixed current angle measurement type error.
    """

    _message = (
        "Mixture of different current angle measurement types on the same terminal for {n} {objects}. "
        "If multiple current sensors measure the same terminal of the same object, all angle measurement types must be "
        "the same. Mixing local_angle and global_angle current measurements on the same terminal is not supported."
    )


class MixedPowerCurrentSensorError(MultiComponentValidationError):
    """
    Mixed power and current sensor error.
    """

    _message = (
        "Mixture of power and current sensors on the same terminal for {n} {objects}. "
        "If multiple sensors measure the same terminal of the same object, all sensors must measure the same quantity."
    )


class MissingVoltageAngleMeasurementError(MultiComponentValidationError):
    """
    Missing voltage angle measurement error.
    """

    _message = (
        "Missing voltage angle measurement for {n} {objects}. "
        "If a voltage sensor measures the voltage of a terminal, it must also measure the voltage angle."
    )



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np

from power_grid_model import (
    ComponentType,
    DatasetType,
    attribute_dtype,
    attribute_empty_value,
    initialize_array,
    power_grid_meta_data,
)


def test_nan_scalar():
    assert np.isnan(power_grid_meta_data[DatasetType.input][ComponentType.node].nan_scalar["u_rated"])


def test_initialize_array():
    arr = initialize_array(DatasetType.input, ComponentType.node, 3)
    assert arr.shape == (3,)
    assert np.all(np.isnan(arr["u_rated"]))
    arr_2d = initialize_array(DatasetType.input, ComponentType.node, (2, 3))
    assert arr_2d.shape == (2, 3)
    assert np.all(np.isnan(arr_2d["u_rated"]))


def test_attribute_dtype():
    assert attribute_dtype(DatasetType.input, ComponentType.node, "u_rated") == np.float64
    assert attribute_dtype(DatasetType.input, ComponentType.node, "id") == np.int32


def test_attribute_empty_value():
    empty_value = attribute_empty_value(DatasetType.input, ComponentType.node, "u_rated")
    assert np.isnan(empty_value)
    empty_value = attribute_empty_value(DatasetType.input, ComponentType.node, "id")
    assert empty_value == np.iinfo(np.int32).min


def test_sensor_meta_data():
    sensors = [
        ComponentType.sym_voltage_sensor,
        ComponentType.asym_voltage_sensor,
        ComponentType.sym_power_sensor,
        ComponentType.asym_power_sensor,
    ]
    input_voltage = ["u_measured", "u_angle_measured", "u_sigma"]
    output_voltage = ["u_residual", "u_angle_residual"]
    input_power = ["p_measured", "q_measured", "power_sigma"]
    output_power = ["p_residual", "q_residual"]
    for sensor in sensors:
        for meta_type in [DatasetType.input, DatasetType.update, DatasetType.sym_output, DatasetType.asym_output]:
            meta_data = power_grid_meta_data[meta_type]
            assert sensor in meta_data
            meta_data_sensor = meta_data[sensor]
            attr_names = meta_data_sensor.dtype_dict["names"]
            assert "id" in attr_names
            # check specific attributes
            if "voltage" in sensor:
                expected_attrs = output_voltage if "output" in meta_type else input_voltage
            else:
                expected_attrs = output_power if "output" in meta_type else input_power

            for name in expected_attrs:
                assert name in attr_names


def test_dict_like_access():
    assert (
        power_grid_meta_data[DatasetType.input][ComponentType.node].dtype
        == power_grid_meta_data[DatasetType.input][ComponentType.node]["dtype"]
    )


def test_all_datasets():
    assert set(power_grid_meta_data.keys()) == {
        "input",
        "update",
        "sym_output",
        "asym_output",
        "sc_output",
        DatasetType.input,
        DatasetType.update,
        DatasetType.sym_output,
        DatasetType.asym_output,
        DatasetType.sc_output,
    }



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

"""
Common enumerations used by the power-grid-model library.

Note: these enumeration match the C++ arithmetic core, so don't change the values unless you change them in C++ as well

"""

from power_grid_model._core.enum import (  # noqa: F401
    AngleMeasurementType,
    Branch3Side,
    BranchSide,
    CalculationMethod,
    CalculationType,
    ComponentAttributeFilterOptions,
    FaultPhase,
    FaultType,
    LoadGenType,
    MeasuredTerminalType,
    ShortCircuitVoltageScaling,
    TapChangingStrategy,
    WindingType,
    _ExperimentalFeatures,
)



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

from pathlib import Path

from power_grid_model import CalculationType
from power_grid_model.utils import import_input_data, import_update_data
from power_grid_model.validation import errors_to_string, validate_batch_data

input_file = Path("../tests/data/power_flow/dummy-test-batch/input.json")
update_file = Path("../tests/data/power_flow/dummy-test-batch/update_batch.json")

input_data = import_input_data(json_file=input_file)
update_data = import_update_data(json_file=update_file)

update_errors = validate_batch_data(
    input_data=input_data, update_data=update_data, calculation_type=CalculationType.power_flow, symmetric=True
)

print(errors_to_string(update_errors, name=str(update_file), details=True))



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import os
import platform
from importlib.resources import files
from pathlib import Path


def get_pgm_dll_path() -> Path:
    """
    Returns the path to PGM dynamic library.
    """
    package_dir = Path(str(files(__package__)))
    platform_name = platform.uname().system
    if platform_name == "Windows":
        lib_dir_1 = package_dir / "bin"
        lib_dir_2 = package_dir / "bin"
    else:
        lib_dir_1 = package_dir / "lib"
        lib_dir_2 = package_dir / "lib64"

    # determine DLL file name
    if platform_name == "Windows":
        dll_file = Path("power_grid_model_c.dll")
    elif platform_name == "Darwin":
        dll_file = Path("libpower_grid_model_c.dylib")
    elif platform.system() == "Linux":
        dll_file = Path("libpower_grid_model_c.so")
    else:
        raise NotImplementedError(f"Unsupported platform: {platform_name}")
    lib_dll_path_1 = lib_dir_1 / dll_file
    lib_dll_path_2 = lib_dir_2 / dll_file

    # determine editable path to the DLL
    # __file__
    #   -> power_grid_model_c (..)
    #     -> _core (..)
    #       -> power_grid_model (..)
    #         -> src (..)
    #           -> repo_root (..)
    #             -> build
    #               -> bin
    editable_dir = Path(__file__).resolve().parent.parent.parent.parent.parent / "build" / "bin"
    editable_dll_path = editable_dir / dll_file

    # first try to load from lib_dll_path
    # then editable_dll_path
    # then if it is inside conda, just load dll_file, the system tries to find it in the PATH
    # then for anything else, raise an error
    if lib_dll_path_1.exists():
        final_dll_path = lib_dll_path_1
    elif lib_dll_path_2.exists():
        final_dll_path = lib_dll_path_2
    elif editable_dll_path.exists():
        final_dll_path = editable_dll_path
    elif os.environ.get("CONDA_PREFIX"):
        final_dll_path = dll_file
    else:
        raise ImportError(f"Could not find shared library: {dll_file}. Your PGM installation may be broken.")

    return final_dll_path



# SPDX-FileCopyrightText: Contributors to the Power Grid Model project <powergridmodel@lfenergy.org>
#
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd

from power_grid_model import ComponentType, DatasetType, LoadGenType, PowerGridModel, initialize_array

"""
node_1 ---line_3--- node_2 ---line_6--- node_7
|                    |                   |
|                    |                   |
source_4          asym_load_5          asym_load_8
"""

# node
node = initialize_array(DatasetType.input, ComponentType.node, 3)
# The following is also supported
# node = initialize_array("input", "node", 3)
node["id"] = [1, 2, 7]
node["u_rated"] = [10.5e3, 10.5e3, 10.5e3]

# line
line = initialize_array(DatasetType.input, ComponentType.line, 2)
line["id"] = [3, 6]
line["from_node"] = [1, 2]
line["to_node"] = [2, 7]
line["from_status"] = [1, 1]
line["to_status"] = [1, 1]
line["r1"] = [0.25, 0.25]
line["x1"] = [0.2, 0.2]
line["c1"] = [10e-6, 10e-6]
line["tan1"] = [0.0, 0.0]
line["r0"] = [0.25, 0.25]  # zero sequence parameters
line["x0"] = [0.2, 0.2]  # zero sequence parameters
line["c0"] = [10e-6, 10e-6]  # zero sequence parameters
line["tan0"] = [0.0, 0.0]  # zero sequence parameters

# load
asym_load = initialize_array(DatasetType.input, ComponentType.asym_load, 2)
asym_load["id"] = [4, 8]
asym_load["node"] = [2, 7]
asym_load["status"] = [1, 1]
asym_load["type"] = [LoadGenType.const_power]

# input for three phase per entry
asym_load["p_specified"] = [[2e6, 0.0, 0.0], [0.0, 1e6, 0.0]]
asym_load["q_specified"] = [[0.5e6, 0.0, 0.0], [0.0, 0.2e6, 0.0]]  # input for three phase per entry

# source
source = initialize_array(DatasetType.input, ComponentType.source, 1)
source["id"] = [5]
source["node"] = [1]
source["status"] = [1]
source["u_ref"] = [1.0]

# input_data
input_data = {
    ComponentType.node: node,
    ComponentType.line: line,
    ComponentType.asym_load: asym_load,
    ComponentType.source: source,
}

# call constructor
model = PowerGridModel(input_data, system_frequency=50.0)

result = model.calculate_power_flow(symmetric=False)

print("Node Input")
print(pd.DataFrame(input_data[ComponentType.node]))
print("Node Result")
print(result[ComponentType.node]["u"])  # N*3 array, in symmetric calculation is this N array
print(result[ComponentType.asym_load]["p"])  # N*3 array, in symmetric calculation is this N array

# batch calculation
scaler = np.linspace(0, 1, 1000)
batch_p = asym_load["p_specified"].reshape(1, 2, 3) * scaler.reshape(-1, 1, 1)
batch_load = initialize_array(DatasetType.update, ComponentType.asym_load, (1000, 2))
batch_load["id"] = [[4, 8]]
batch_load["p_specified"] = batch_p
batch_update = {ComponentType.asym_load: batch_load}

result = model.calculate_power_flow(symmetric=False, update_data=batch_update)
print(result[ComponentType.node]["u"].shape)  # 1000 (scenarios) *3 (nodes) *3 (phases)
